/*
 * FlowImplementation.hpp
 */

#ifndef FLOW_IMPLEMENTATION_HPP_
#define FLOW_IMPLEMENTATION_HPP_

#include <string>
#include <vector>
#include <algorithm>
//#include <cassert>
#include <iostream>
#include <queue>
#include <limits>
#include "Parser.hpp"
#include "bucket_queue.hpp"

using namespace std;

class FlowGraph{
  private:
    FlowGraph( const FlowGraph & ){} //do not copy
    void operator=( const FlowGraph & ){} //really, do not copy

    static bool comp(Parser::Arc a1, Parser::Arc a2) {
      if (a1.source < a2.source)
        return true;

      if (a1.source > a2.source)
        return false;

      if (a1.target < a2.target)
        return true;

      return false;
    }

    static bool comp2(Parser::Arc a1, Parser::Arc a2) {
      if (a1.target < a2.target)
        return true;

      if (a1.target > a2.target)
        return false;

      if (a1.source < a2.source)
        return true;

      return false;
    }

  public:
    struct Node {
      unsigned int firstEdge;
      unsigned int currentEdge;
      unsigned int distance;
      unsigned int excess;
    };

    struct Edge {
      unsigned int target;
      unsigned int capacity;
      unsigned int flow;
      bool reverse;
    };

    Node* v;
    Edge* e;

    unsigned int n;
    unsigned int m;

    FlowGraph() {
    }

    FlowGraph( const Parser & parser ) {
      n = parser.num_nodes;
      //m = parser.arcs.size();
      v = new Node[n + 2]; // begins at 1
      e = new Edge[2 * parser.arcs.size()];

      // TODO: Darf ich das hier schon?
      //v[parser.nodes[0]].distance = n;

      //v[parser.num_nodes + 1].firstEdge = e; // guard

      vector<Parser::Arc> sortedArcs = parser.arcs;
      sort(sortedArcs.begin(), sortedArcs.end(), comp);

      vector<Parser::Arc> rSortedArcs = parser.arcs;
      sort(rSortedArcs.begin(), rSortedArcs.end(), comp2);

      unsigned int currentNode = 1;
      unsigned int currentEdge = 0;

      v[currentNode].firstEdge = currentEdge;
      v[currentNode].currentEdge = currentEdge;

      unsigned int si = 0;
      unsigned int ri = 0;

      while (si != sortedArcs.size() || ri != rSortedArcs.size()) {
        if (si == sortedArcs.size() || sortedArcs.at(si).source > rSortedArcs.at(ri).target) {
          Parser::Arc& arc = rSortedArcs.at(ri);

          if (arc.target != currentNode) {
            currentNode = arc.target;
            v[currentNode].firstEdge = currentEdge;
            v[currentNode].currentEdge = currentEdge;
          }

          e[currentEdge].target = arc.source;
          e[currentEdge].capacity = 0;
          e[currentEdge].reverse = true;

          ri++;
        } else {
          Parser::Arc& arc = sortedArcs.at(si);

          if (arc.source != currentNode) {
            currentNode = arc.source;
            v[currentNode].firstEdge = currentEdge;
            v[currentNode].currentEdge = currentEdge;
          }

          e[currentEdge].target = arc.target;
          e[currentEdge].capacity = arc.capacity;
          e[currentEdge].reverse = false;

          si++;
        }

        currentEdge++;
      }

      v[currentNode + 1].firstEdge = currentEdge; // guard
    }
};

class FlowImplementation{
private:
  //FlowImplementation( const FlowImplementation& ){} //do not copy
  //void operator=( const FlowImplementation& ){} //really, do not copy

  FlowGraph graph;
  //queue<unsigned int> fifo;
  PriorityQueue<unsigned int, unsigned int, std::numeric_limits<unsigned int> > q;

  unsigned int getMask(unsigned int n) const {
    n += 2;

    // to next power of 2 ...
    unsigned int mask = 1;

    while (n >>= 1) {
      mask++;
    }

    return (1 << mask) - 1; // ... and - 1
  }


public:
  FlowImplementation( const Parser & parser )
    : graph( parser )
    , q(graph.n, getMask(graph.n))
  {
    q.clear();
  }

  void push(FlowGraph::Edge& e, unsigned int v, unsigned int d) {
    //cout << v << ", " << e.target << ", " << d << ", " << graph.v[v].excess << endl;
    //assert(d > 0);
    //assert(graph.v[v].excess >= d);

    unsigned int i = graph.v[e.target].firstEdge;

    while (graph.e[i].target != v || graph.e[i].reverse == e.reverse) {
      i++;
    }

    //assert(i < graph.v[e.target + 1].firstEdge);

    FlowGraph::Edge& re = graph.e[i];

    e.flow += d;
    re.flow -= d;

    e.capacity -= d;
    re.capacity += d;

    graph.v[v].excess -= d;
    graph.v[e.target].excess += d;
  }

  //berechnet den Wert des Flusses zwischen Quelle (source) und Senke( sink )
  unsigned int calculateFlow( unsigned int source, unsigned int sink ) {
    for (unsigned int i = 0; i <= graph.n; i++) {
      graph.v[i].distance = 0;
      graph.v[i].excess = 0;
    }
    //cout << endl << graph.n << endl;
    graph.v[source].distance = graph.n;
    graph.v[source].excess = -1; // unnecessary

    for (unsigned int i = graph.v[source].firstEdge; i < graph.v[source + 1].firstEdge; ++i) {
      if (!graph.e[i].reverse) {
        push(graph.e[i], source, graph.e[i].capacity);
        q.push(graph.e[i].target, graph.v[graph.e[i].target].distance);
      }
    }

    //cout << "saturated" << endl;

    // naive approach
    bool wasExcess;

    do {
      wasExcess = false;

      unsigned int v;
      while (!q.empty()) {
        v = q.getMin();
        q.deleteMin();

        if (v == source || v == sink) {
          continue;
        }

        if (graph.v[v].excess == 0) {
          continue;
        }

        wasExcess = true;

        unsigned int ce = graph.v[v].currentEdge;
        unsigned int fe = graph.v[v].firstEdge;
        unsigned int startPos = ce - fe;
        unsigned int w;

        bool found = false;

        for (unsigned int j = ce; j < graph.v[v+1].firstEdge; j++) {
          w = graph.e[j].target;

          if (graph.v[v].excess == 0) {
            break;
          }

          if (graph.e[j].capacity == 0) {
            continue;
          }

          if (!q.contains(w)) {
            q.push(w, graph.v[w].distance); // maybe move down
          } else {
            q.decreaseKey(w, graph.v[w].distance);
          }

          if (graph.v[w].distance < graph.v[v].distance) {
            found = true;
            bool isSaturating = graph.e[j].capacity <= graph.v[v].excess;
            push(graph.e[j], v, min(graph.v[v].excess, graph.e[j].capacity));
            graph.v[v].currentEdge++;

            if (graph.v[v].currentEdge >= graph.v[v+1].firstEdge) {
              graph.v[v].currentEdge = graph.v[v].firstEdge;
            }

            if (isSaturating) {
              break;
            }
          }
        }

        if (!found) {
          graph.v[v].distance++;
          graph.v[v].currentEdge = graph.v[v].firstEdge;
          if (!q.contains(v)) {
            q.push(v, graph.v[v].distance);
          } else {
            q.decreaseKey(v, graph.v[v].distance);
          }
        }
      }
    } while (wasExcess);

    //cout << "flow calculated" << endl;

    unsigned int flow = 0;
    for (unsigned int i = graph.v[source].firstEdge; i < graph.v[source + 1].firstEdge; i++) {
      flow += graph.e[i].flow;
    }

    return flow;
  }
};


#endif /* FLOW_IMPLEMENTATION_HPP_ */
